@@ -15,7 +15,9 @@
 --
 --   $ lumail --rcfile ./one.lua --rcfile ./two.lua
 --
--- NOTE: This file will also cause ~/.lumail.d/*.lua to be loaded.
+-- NOTE: This file will also load anything matching the pattern
+-- ~/.lumail.d/*.lua - This is implemented via a call to the "load_directory"
+-- primitive, which is one of the primitives that Lumail supplies.
 --
 ---
 ---
@@ -51,7 +53,7 @@
 prefix = os.getenv( "HOME" ) .. "/Maildir"
 
 --
--- Ensure the directory exists.
+-- Ensure the directory exists, and abort with an error if it doesn't.
 --
 if ( is_directory(prefix) ) then
    maildir_prefix(prefix)
@@ -79,7 +81,7 @@ end
 -- This can be changed in one of the available hooks.
 --
 -- In this configuration file we use the `on_folder_selection()` function
--- to change this on a per-folder basis.
+-- later on to change this on a per-folder basis.
 --
 default_email = "Steve Kemp <steve@steve.org.uk>"
 from( default_email )
@@ -122,7 +124,7 @@ end
 --
 -- When called with no arguments this function will return the current
 -- setting, otherwise it will update it.  This is a common pattern for lumail,
--- each of the functions which can get or set a value works in the same way.
+-- each of the functions which can get/set a value works in the same way.
 --
 -- The maildir limit can be set to three values:
 --
@@ -130,7 +132,7 @@ end
 --    new  -> Show all maildir folders which contain unread messages.
 --   "pat" -> Show all maildir folders which match the regular expression "pat".
 --
--- See "all_folders" , "new_folders", and "livejournal_folders" functions for
+-- See "all_folders", "new_folders", and "livejournal_folders" functions for
 -- example of use.
 --
 maildir_limit( "all" )
@@ -186,7 +188,7 @@ index_format( "[$FLAGS] $DAY/$MONTH/$YEAR $FROM - $SUBJECT" )
 --
 -- The headers we show when displaying a mail.
 --
--- (The headers listed are shown in the order they are specified here.)
+-- (The headers displayed are presented in the order they are specified here.)
 --
 headers = { "$TO", "$CC", "$FROM", "$DATE", "$SUBJECT" }
 
@@ -244,6 +246,7 @@ function offlineimap()
    return true
 end
 
+
 --
 -- This function is called when the client is launched.
 --
@@ -254,15 +257,15 @@ end
 --    end
 --
 function on_start()
-   msg("lumail v" .. VERSION .. " http://lumail.org/" )
+   show_version()
 end
 
 
 --
--- This function is called when a folder is added/removed to the selected set.,
--- or toggled.
+-- This function is called when a folder is added/removed/toggled
+-- within the selected set.
 --
--- It is where you can run per-folder hooks.
+-- This function is a perfect place to run per-folder hooks.
 --
 -- Remember: More than one folder might be selected at any given time.
 --
@@ -279,7 +282,7 @@ function on_folder_selection( folder )
       from( default_email )
    end
 
-   msg( "There are " .. count_messages() .. " messages" )
+   msg( "There are " .. count_messages() .. " messages present." )
 end
 
 
@@ -302,9 +305,12 @@ function jump_to_end()
       jump_index_to( count_messages() - 1 )
 
    else
+      --
+      -- TODO: Fix this for messages.
+      --
 
       -- Jump to the end of the message.
-      scroll_message_to( count_lines() - 2 )
+      jump_message_to( count_lines() - 2 )
    end
 end
 
@@ -328,9 +334,12 @@ function jump_to_start()
       jump_index_to( 0 )
 
    else
+      --
+      -- TODO: Fix this for message/text
+      --
 
       -- Jump to the start of the message.
-      scroll_message_to( 0 )
+      jump_message_to( 0 )
    end
 end
 
@@ -493,18 +502,37 @@ end
 --
 -- This function will be called when the global mode changes
 --
-function on_mode_change( old, new)
-    -- check if we are switching from maildir to index mode, if so try to jump
-    -- the first unread mail or go to the last one
-    if string.find(old,'maildir') and string.find(new,'index') then
-        if (newmail_displayed()) then
+-- Here we define a closure so that the previous mode can be
+-- returned to easily.
+--
+-- (In our case this is so that when viewing a file you can revert
+-- to the previous mode.  See the `display_file` primitive and the
+-- function named 'show_file` later in this file.)
+--
+do
+
+   -- We store the previous global-mode here, when changing the mode.
+   local pmode = ''
+
+   function on_mode_change( old, new )
+      -- check if we are switching from maildir to index mode, if so try to jump
+      -- the first unread mail or go to the last one
+      if string.find(old,'maildir') and string.find(new,'index') then
+         if (newmail_displayed()) then
             jump_to_first_unread()
-        else
+         else
             jump_to_end()
-        end
-    end
+         end
+      end
+      pmode = old
+   end
+
+   function previous_mode()
+      global_mode( pmode )
+   end
 end
 
+
 --
 -- Test if there is currently a new message displayed
 --
@@ -594,6 +622,16 @@ function view_message()
 end
 
 
+--
+-- Show a file.
+--
+function show_file( filename )
+   show_file_contents(filename)
+   global_mode( "text" )
+   clear()
+end
+
+
 --
 -- When in maildir-mode show all folders.
 --
@@ -623,44 +661,93 @@ end
 
 
 --
--- Navigation functions for maildir-mode
+-- Navigation function:  Scroll "something" down.
 --
-function maildir_down()
-   scroll_maildir_down( 1 )
+function down()
+   m = global_mode()
+
+   if ( string.find( m, "maildir" ) ) then
+      scroll_maildir_down( 1 )
+   elseif (string.find(m, "index" ) ) then
+      scroll_index_down( 1 )
+   elseif (string.find(m, 'message') ) then
+      scroll_message_down( 1 )
+   elseif (string.find(m, 'text') ) then
+      scroll_text_down( 1 )
+   else
+      msg( "down() not implemented for mode:" .. m )
+   end
 end
-function maildir_page_down()
-   -- The minus-two is to account for the status-area
-   scroll_maildir_down( screen_height() - 2  )
-end
-function maildir_up()
-   scroll_maildir_up(1)
-end
-function maildir_page_up()
-   -- The minus-two is to account for the status-area
-   scroll_maildir_up( screen_height() -2 )
+
+--
+-- Navigation function:  Scroll "something" up.
+--
+function up()
+   m = global_mode()
+
+   if ( string.find( m, "maildir" ) ) then
+      scroll_maildir_up( 1 )
+   elseif (string.find(m, "index" ) ) then
+      scroll_index_up( 1 )
+   elseif (string.find(m, 'message') ) then
+      scroll_message_up( 1 )
+   elseif (string.find(m, 'text') ) then
+      scroll_text_up( 1 )
+   else
+      msg( "up() not implemented for mode:" .. m )
+   end
 end
 
 
+--
+-- Navigation function:  Scroll "something" down a page.
+--
+function page_down()
+   m = global_mode()
+
+   if ( string.find( m, "maildir" ) ) then
+      -- The minus-two is to account for the status-area
+      scroll_maildir_down( screen_height() - 2  )
+   elseif (string.find(m, "index" ) ) then
+      -- The minus-two is to account for the status-area
+      scroll_index_down( screen_height() - 2  )
+   elseif (string.find(m, 'message') ) then
+      -- The minus-two is to account for the status-area
+      scroll_message_down( screen_height() - 2  )
+   elseif (string.find(m, 'text') ) then
+      -- The minus-two is to account for the status-area
+      scroll_text_down( screen_height() - 2  )
+   else
+      msg( "page_down() not implemented for mode:" .. m )
+   end
+end
+
 
 --
--- Navigation functions for index-mode
+-- Navigation function:  Scroll "something" up a page.
 --
-function message_down()
-   scroll_index_down( 1 )
-end
-function message_page_down()
-   -- The minus-two is to account for the status-area
-   scroll_index_down( screen_height() - 2 )
-end
-function message_up()
-   scroll_index_up(1)
-end
-function message_page_up()
-   -- The minus-two is to account for the status-area
-   scroll_index_up( screen_height() - 2 )
+function page_up()
+   m = global_mode()
+
+   if ( string.find( m, "maildir" ) ) then
+      -- The minus-two is to account for the status-area
+      scroll_maildir_up( screen_height() - 2  )
+   elseif (string.find(m, "index" ) ) then
+      -- The minus-two is to account for the status-area
+      scroll_index_up( screen_height() - 2  )
+   elseif (string.find(m, 'message') ) then
+      -- The minus-two is to account for the status-area
+      scroll_message_up( screen_height() - 2  )
+   elseif (string.find(m, 'text') ) then
+      -- The minus-two is to account for the status-area
+      scroll_text_up( screen_height() - 2  )
+   else
+      msg( "page_up() not implemented for mode:" .. m )
+   end
 end
 
 
+
 --
 -- Should we wrap lines when showing a message?
 --
@@ -685,7 +772,7 @@ do
    local search_next_prev = ''
 
    --
-   -- Search for the next folder/message which matches the entered pattern.
+   -- Search forwards given the user-specified pattern.
    --
    function search_next()
 
@@ -705,7 +792,7 @@ do
       search_next_prev = x
 
       --
-      --  Get the mode, so we can do the correct kind of searching.
+      --  Get the mode, so we can perform the correct type of search.
       --
       m = global_mode()
 
@@ -713,6 +800,10 @@ do
          scroll_maildir_to( x )
       elseif (string.find(m, "index" ) ) then
          scroll_index_to( x )
+      elseif (string.find(m, "message" ) ) then
+         scroll_message_to( x )
+      elseif (string.find(m, "text" ) ) then
+         scroll_text_to( x )
       else
          msg( "search_next() not implemented for mode:" .. m )
       end
@@ -724,17 +815,24 @@ end
 --
 -- Open a folder exclusively.
 --
--- This behaves like the traditional "open folder" function in our favourite
--- client.  I like to open multiple folders at once, because that's how I roll.
+-- If called with no arguments it will open the folder under the
+-- cursor in maildir-mode.
 --
-function open_folder()
+-- If called with the name of a folder it will open that specified
+-- folder.
+--
+function open_folder(folder)
    clear_selected_folders()
-   add_selected_folder()
+
+   if ( folder ) then
+      set_selected_folder(folder)
+   else
+      add_selected_folder()
+   end
    index()
 end
 
 
-
 --
 -- Open my most important folders.
 --
@@ -921,13 +1019,15 @@ keymap['global']  = {}
 keymap['index']   = {}
 keymap['message'] = {}
 keymap['maildir'] = {}
-
+keymap['text']    = {}
 
 --
 -- Global Bindings
 --
 -- Quit
 keymap['global']['Q'] = "exit()"
+-- Help
+keymap['global']['H'] = "help()"
 -- Compose a new message.
 keymap['global']['c'] = "compose()"
 -- repl-mode
@@ -940,15 +1040,31 @@ keymap['global']['M']   = 'maildir()'
 keymap['global']['I'] = 'index()'
 -- Show the client version
 keymap['global']['v'] = 'show_version()'
+
 -- Jump to end
 keymap['global']['*']        = 'jump_to_end()'
 keymap['global']['KEY_END']  = 'jump_to_end()'
 keymap['global']['KEY_HOME'] = 'jump_to_start()'
+
 -- Refresh display
 keymap['global']['^R'] = 'clear()'
 keymap['global']['^L'] = 'clear()'
 keymap['global']['KEY_RESIZE'] = 'clear()'
 
+-- Scroll up/down
+keymap['global']['j'] = 'down()'
+keymap['global']['KEY_DOWN'] = 'down()'
+keymap['global']['J'] = 'page_down()'
+keymap['global']['KEY_NPAGE'] = 'page_down()'
+
+keymap['global']['k'] = 'up()'
+keymap['global']['KEY_UP'] = 'up()'
+keymap['global']['K'] = 'page_up()'
+keymap['global']['KEY_PPAGE'] = 'page_up()'
+
+-- Search forwards, by regular expression
+keymap['global']['/'] = 'search_next()'
+
 
 --
 -- THIS WORKS :D.
@@ -977,33 +1093,14 @@ keymap['maildir']['g'] = 'maildir_limit( "(Google.*INBOX|Google.*All Mail)" )'
 
 
 --
--- Scroll up/down & find folders
+-- Find folders
 --
-keymap['maildir']['j'] = 'maildir_down()'
-keymap['maildir']['KEY_DOWN'] = 'maildir_down()'
-keymap['maildir']['J'] = 'maildir_page_down()'
-keymap['maildir']['KEY_NPAGE'] = 'maildir_page_down()'
-keymap['maildir']['k'] = 'maildir_up()'
-keymap['maildir']['KEY_UP'] = 'maildir_up()'
-keymap['maildir']['K'] = 'maildir_page_up()'
-keymap['maildir']['KEY_PPAGE'] = 'maildir_page_up()'
-keymap['maildir']['/'] = 'search_next()'
 keymap['maildir']['f'] = 'faves()'
 
 
 --
--- Scroll up/down the message index
+-- Per-message actions
 --
-keymap['index']['j'] = 'message_down()'
-keymap['index']['KEY_DOWN'] = 'message_down()'
-keymap['index']['J'] = 'message_page_down()'
-keymap['index']['KEY_NPAGE'] = 'message_page_down()'
-keymap['index']['k'] = 'message_up()'
-keymap['index']['KEY_UP'] = 'message_up()'
-keymap['index']['K'] = 'message_page_up()'
-keymap['index']['KEY_PPAGE'] = 'message_page_up()'
-keymap['index']['/'] = 'search_next()'
-
 keymap['index']['Space'] = 'view_message()'
 keymap['index']['Enter'] = 'view_message()'
 keymap['index']['r'] = 'reply()'
@@ -1019,7 +1116,7 @@ keymap['index']['n'] = 'index_limit("new")'
 --
 -- Jump to the next unread mail in index or message mode
 --
-keymap['index']['N'] = 'jump_to_next_unread()'
+keymap['index']['N']   = 'jump_to_next_unread()'
 keymap['message']['N'] = 'jump_to_next_unread()'
 
 --
@@ -1037,20 +1134,6 @@ keymap['maildir']['Enter'] = 'open_folder()'
 -- Message-handling.
 --
 --
--- Scroll up/down in the current message.
-keymap['message']['j'] = 'scroll_message_down(1)'
-keymap['message']['KEY_DOWN'] = 'scroll_message_down(1)'
-keymap['message']['KEY_NPAGE'] = 'scroll_message_down(10)'
-keymap['message']['k'] = 'scroll_message_up(1)'
-keymap['message']['KEY_UP'] = 'scroll_message_up(1)'
-keymap['message']['KEY_PPAGE'] = 'scroll_message_up(10)'
-keymap['message']['Space'] = 'scroll_message_down(10)'
-
---
--- scroll to the next/prev message
---
-keymap['message']['J'] = 'message_down()'
-keymap['message']['K'] = 'message_up()'
 
 -- wrap linese
 keymap['message']['w'] = 'toggle_wrap_lines()'
@@ -1060,6 +1143,16 @@ keymap['message']['d'] = 'delete()'
 keymap['message']['f'] = 'forward()'
 keymap['message']['r'] = 'reply()'
 
+-- Override the navigation of next/previous message
+-- which would otherwise scroll the current message text down/up by a page
+keymap['message']['J']     = 'scroll_index_down(1)'
+keymap['message']['K']     = 'scroll_index_up(1)'
+keymap['message']['Space'] = "page_down()"
+
+--
+-- Text display
+--
+keymap['text']['q'] = 'previous_mode()'
 
 
 ---
